@@ -92,11 +92,6 @@ Permutations, Tuples, Integer sequences
 .. autofunction:: generate_permutations
 .. autofunction:: generate_unique_permutations
 
-Graph Algorithms
-----------------
-
-.. autofunction:: a_star
-
 Formatting
 ----------
 
@@ -136,6 +131,11 @@ Helpers for :mod:`numpy`
 Timing data
 -----------
 
+.. data:: SUPPORTS_PROCESS_TIME
+
+   A :class:`bool` indicating whether :class:`ProcessTimer` measures elapsed
+   process time (available on Python 3.3+).
+
 .. autoclass:: ProcessTimer
 
 Log utilities
@@ -144,6 +144,12 @@ Log utilities
 .. autoclass:: ProcessLogger
 .. autoclass:: DebugProcessLogger
 .. autoclass:: log_process
+
+Sorting in natural order
+------------------------
+
+.. autofunction:: natorder
+.. autofunction:: natsorted
 """
 
 
@@ -315,12 +321,18 @@ class Record(RecordWithoutPickling):
 
 class ImmutableRecordWithoutPickling(RecordWithoutPickling):
     "Hashable record. Does not explicitly enforce immutability."
+    def __init__(self, *args, **kwargs):
+        RecordWithoutPickling.__init__(self, *args, **kwargs)
+        self._cached_hash = None
 
     def __hash__(self):
-        return hash(
+        if self._cached_hash is None:
+            self._cached_hash = hash(
                 (type(self),) + tuple(getattr(self, field)
                     for field in self.__class__.fields))
 
+        return self._cached_hash
+
 
 class ImmutableRecord(ImmutableRecordWithoutPickling, Record):
     pass
@@ -1034,10 +1046,15 @@ def argmax(iterable):
 
 # {{{ cartesian products etc.
 
-def cartesian_product(list1, list2):
-    for i in list1:
-        for j in list2:
-            yield (i, j)
+def cartesian_product(*args):
+    if len(args) == 1:
+        for arg in args[0]:
+            yield (arg,)
+        return
+    first = args[:-1]
+    for prod in cartesian_product(*first):
+        for i in args[-1]:
+            yield prod + (i,)
 
 
 def distinct_pairs(list1, list2):
@@ -1347,68 +1364,27 @@ def get_write_to_map_from_permutation(original, permuted):
 # }}}
 
 
+# {{{ code maintenance
+
+class MovedFunctionDeprecationWrapper:
+    def __init__(self, f):
+        self.f = f
+
+    def __call__(self, *args, **kwargs):
+        from warnings import warn
+        warn("This function is deprecated. Use %s.%s instead." % (
+            self.f.__module__, self.f.__name__),
+            DeprecationWarning, stacklevel=2)
+
+        return self.f(*args, **kwargs)
+
+# }}}
+
+
 # {{{ graph algorithms
 
-def a_star(  # pylint: disable=too-many-locals
-        initial_state, goal_state, neighbor_map,
-        estimate_remaining_cost=None,
-        get_step_cost=lambda x, y: 1
-        ):
-    """
-    With the default cost and heuristic, this amounts to Dijkstra's algorithm.
-    """
-
-    from heapq import heappop, heappush
-
-    if estimate_remaining_cost is None:
-        def estimate_remaining_cost(x):  # pylint: disable=function-redefined
-            if x != goal_state:
-                return 1
-            else:
-                return 0
-
-    class AStarNode(object):
-        __slots__ = ["state", "parent", "path_cost"]
-
-        def __init__(self, state, parent, path_cost):
-            self.state = state
-            self.parent = parent
-            self.path_cost = path_cost
-
-    inf = float("inf")
-    init_remcost = estimate_remaining_cost(initial_state)
-    assert init_remcost != inf
-
-    queue = [(init_remcost, AStarNode(initial_state, parent=None, path_cost=0))]
-    visited_states = set()
-
-    while queue:
-        _, top = heappop(queue)
-        visited_states.add(top.state)
-
-        if top.state == goal_state:
-            result = []
-            it = top
-            while it is not None:
-                result.append(it.state)
-                it = it.parent
-            return result[::-1]
-
-        for state in neighbor_map[top.state]:
-            if state in visited_states:
-                continue
-
-            remaining_cost = estimate_remaining_cost(state)
-            if remaining_cost == inf:
-                continue
-            step_cost = get_step_cost(top, state)
-
-            estimated_path_cost = top.path_cost+step_cost+remaining_cost
-            heappush(queue,
-                (estimated_path_cost,
-                    AStarNode(state, top, path_cost=top.path_cost + step_cost)))
-
-    raise RuntimeError("no solution")
+from pytools.graph import a_star as a_star_moved
+a_star = MovedFunctionDeprecationWrapper(a_star_moved)
 
 # }}}
 
@@ -1436,12 +1412,12 @@ class Table:
         col_widths = [max(len(row[i]) for row in self.rows)
                       for i in range(columns)]
 
-        lines = [
-            "|".join([cell.ljust(col_width)
-                      for cell, col_width in zip(row, col_widths)])
+        lines = [" | ".join([cell.ljust(col_width)
+            for cell, col_width in zip(row, col_widths)])
             for row in self.rows]
-        lines[1:1] = ["+".join("-"*col_width
-                              for col_width in col_widths)]
+        lines[1:1] = ["+".join("-" * (col_width + 1 + (i > 0))
+            for i, col_width in enumerate(col_widths))]
+
         return "\n".join(lines)
 
     def latex(self, skip_lines=0, hline_after=None):
@@ -1617,23 +1593,6 @@ class CPyUserInterface(object):
 # }}}
 
 
-# {{{ code maintenance
-
-class MovedFunctionDeprecationWrapper:
-    def __init__(self, f):
-        self.f = f
-
-    def __call__(self, *args, **kwargs):
-        from warnings import warn
-        warn("This function is deprecated. Use %s.%s instead." % (
-            self.f.__module__, self.f.__name__),
-            DeprecationWarning, stacklevel=2)
-
-        return self.f(*args, **kwargs)
-
-# }}}
-
-
 # {{{ debugging
 
 class StderrToStdout(object):
@@ -2077,6 +2036,9 @@ def reshaped_view(a, newshape):
 
 # {{{ process timer
 
+SUPPORTS_PROCESS_TIME = (sys.version_info >= (3, 3))
+
+
 class ProcessTimer(object):
     """Measures elapsed wall time and process time.
 
@@ -2096,7 +2058,7 @@ class ProcessTimer(object):
 
     def __init__(self):
         import time
-        if sys.version_info >= (3, 3):
+        if SUPPORTS_PROCESS_TIME:
             self.perf_counter_start = time.perf_counter()
             self.process_time_start = time.process_time()
 
@@ -2114,7 +2076,7 @@ class ProcessTimer(object):
         # pylint: disable=attribute-defined-outside-init
 
         import time
-        if sys.version_info >= (3, 3):
+        if SUPPORTS_PROCESS_TIME:
             self.wall_elapsed = time.perf_counter() - self.perf_counter_start
             self.process_elapsed = time.process_time() - self.process_time_start
 
@@ -2156,9 +2118,39 @@ class ProcessLogger(object):  # pylint: disable=too-many-instance-attributes
 
         import threading
         self.late_start_log_thread = threading.Thread(target=self._log_start_if_long)
+
         # Do not delay interpreter exit if thread not finished.
         self.late_start_log_thread.daemon = True
-        self.late_start_log_thread.start()
+
+        # https://github.com/firedrakeproject/firedrake/issues/1422
+        # Starting a thread may irrecoverably break various environments,
+        # e.g. MPI.
+        #
+        # Since the late-start logging is an optional 'quality-of-life'
+        # feature for interactive use, do not do it unless there is (weak)
+        # evidence of interactive use.
+        import sys
+        if sys.stdin is None:
+            # Can happen, e.g., if pudb is controlling the console.
+            use_late_start_logging = False
+        else:
+            use_late_start_logging = sys.stdin.isatty()
+
+        import os
+        if os.environ.get("PYTOOLS_LOG_NO_THREADS", ""):
+            use_late_start_logging = False
+
+        if use_late_start_logging:
+            try:
+                self.late_start_log_thread.start()
+            except RuntimeError:
+                # https://github.com/firedrakeproject/firedrake/issues/1422
+                #
+                # Starting a thread may fail in various environments, e.g. MPI.
+                # Since the late-start logging is an optional 'quality-of-life'
+                # feature for interactive use, tolerate failures of it without
+                # warning.
+                pass
 
         self.timer = ProcessTimer()
 
@@ -2235,6 +2227,59 @@ class log_process(object):  # noqa: N801
 # }}}
 
 
+# {{{ sorting in natural order
+
+def natorder(item):
+    """Return a key for natural order string comparison.
+
+    See :func:`natsorted`.
+
+    .. versionadded:: 2020.1
+    """
+    import re
+    result = []
+    for (int_val, string_val) in re.findall(r"(\d+)|(\D+)", item):
+        if int_val:
+            result.append(int(int_val))
+            # Tie-breaker in case of leading zeros in *int_val*.  Longer values
+            # compare smaller to preserve order of numbers in decimal notation,
+            # e.g., "1.001" < "1.01"
+            # (cf. https://github.com/sourcefrog/natsort)
+            result.append(-len(int_val))
+        else:
+            result.append(string_val)
+    return result
+
+
+def natsorted(iterable, key=None, reverse=False):
+    """Sort using natural order [1]_, as opposed to lexicographic order.
+
+    Example::
+
+        >>> sorted(["_10", "_1", "_9"]) == ["_1", "_10", "_9"]
+        True
+        >>> natsorted(["_10", "_1", "_9"]) == ["_1", "_9", "_10"]
+        True
+
+    :arg iterable: an iterable to be sorted. It must only have strings, unless
+        *key* is specified.
+    :arg key: if provided, a key function that returns strings for ordering
+        using natural order.
+    :arg reverse: if *True*, sorts in descending order.
+
+    :returns: a sorted list
+
+    .. [1] https://en.wikipedia.org/wiki/Natural_sort_order
+
+    .. versionadded:: 2020.1
+    """
+    if key is None:
+        key = lambda x: x
+    return sorted(iterable, key=lambda y: natorder(key(y)), reverse=reverse)
+
+# }}}
+
+
 def _test():
     import doctest
     doctest.testmod()
